# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import re
import time
import shutil
import typing as typ
import logging
import pathlib as pl

import jinja2

try:
    import importlib.resources as importlib_resources
except ImportError:
    import importlib_resources  # type: ignore

from . import vcs
from . import parse
from . import md2html
from . import html2pdf
from . import __version__
from . import pdf_booklet
from . import vcs_timeline
from . import html_postproc

logger = logging.getLogger("litprog.gen_docs")

MarkdownText = str

HTMLText = str


class Replacement(typ.NamedTuple):

    target  : str
    filename: str


PRINT_FORMATS = (
    'print_ereader',
    'print_a4',
    'print_a5',
    'print_booklet_a4',
    'print_halfletter',
    'print_booklet_letter',
    # 'print_tallcol_a4',
    # 'print_tallcol_letter',
    # 'print_twocol_a4',
    # 'print_twocol_letter',
)

MULTIPAGE_FORMATS = {
    'print_booklet_a4'    : "print_a5",
    'print_booklet_letter': "print_halfletter",
    'print_twocol_a4'     : "print_tallcol_a4",
    'print_twocol_letter' : "print_tallcol_letter",
}

PART_PAGE_SIZES = {
    'screen'              : "screen",
    'print_ereader'       : "ereader",
    'print_a5'            : "a5",
    'print_a4'            : "a4",
    'print_booklet_a4'    : "a5",
    'print_letter'        : "letter",
    'print_halfletter'    : "halfletter",
    'print_booklet_letter': "halfletter",
    'print_tallcol_a4'    : "tallcol_a4",
    'print_tallcol_letter': "tallcol_letter",
    'print_twocol_a4'     : "tallcol_a4",
    'print_twocol_letter' : "tallcol_letter",
}

SELECTED_STATIC_DEPS = {
    r"static/fonts_screen\.css",
    r"static/fonts_print\.css",
    r"static/katex\.css",
    r"static/codehilite\.css",
    r"static/general_v2\.css",
    # r"static/screen_v2\.css",
    r"static/screen_v3\.css",
    # r"static/slideout.js",
    r"static/popper.js",
    # r"static/popper.min.js",
    r"static/app.js",
    r"static/print\.css",
    r"static/print_a4\.css",
    r"static/print_a5\.css",
    r"static/print_letter\.css",
    r"static/print_ereader\.css",
    r"static/print_halfletter\.css",
    r"static/print_tallcol\.css",
    r"static/print_tallcol_a4\.css",
    r"static/print_tallcol_letter\.css",
    r"static/.+\.css",
    r"static/.+\.svg",
    r"static/fonts/.+\.woff2",
    r"static/fonts/.+\.woff",
    r"static/fonts/.+\.ttf",
}


INDEX_HTML = """
<!DOCTYPE html>
<html>
<head> <meta http-equiv="refresh" content="0; URL='{}'" /> </head>
<body></body>
</html>
"""


Metadata = typ.Dict[str, typ.Any]

METADATA_DEFAULTS = {'lang': "en-US", 'title': "-"}


class FileItem(typ.NamedTuple):
    md_path : pl.Path
    meta    : Metadata
    html_res: md2html.HTMLResult

    block_linenos: typ.List[typ.Tuple[int, int]]


FOOTNOTES_RE = re.compile(r'\<h\d\s+id="references"\s*\>')


def wrap_content_html(
    content_html: HTMLText,
    target      : str,
    meta        : Metadata,
    htmls       : typ.Optional[html_postproc.HTMLTexts] = None,
) -> HTMLText:
    assert target == 'screen' or target.startswith('print_')
    meta['target'] = target

    fmt = {
        'page_size'        : PART_PAGE_SIZES[target],
        'is_print_target'  : target.startswith('print_'),
        'is_web_target'    : not target.startswith('print_'),
        'is_tallcol_target': "_tallcol_" in target,
    }

    nav = {}
    if htmls:
        if htmls.chapters:
            nav['chapters_html'] = htmls.chapters
        if htmls.sections:
            nav['sections_html'] = htmls.sections
        if htmls.chapter_prev_href:
            nav['chapter_prev_href'] = htmls.chapter_prev_href
            nav['chapter_prev_text'] = htmls.chapter_prev_text
        if htmls.chapter_next_href:
            nav['chapter_next_href'] = htmls.chapter_next_href
            nav['chapter_next_text'] = htmls.chapter_next_text

    ctx = {'meta': meta, 'fmt': fmt, 'nav': nav, 'content': content_html}

    env    = jinja2.Environment(loader=jinja2.PackageLoader('litprog', package_path="static"))
    tmpl   = env.get_template("template_v2.html")
    result = tmpl.render(**ctx)
    return result


def _deep_update(src: typ.Dict, dest: typ.Dict) -> None:
    for key, src_v in src.items():
        if isinstance(src_v, dict):
            dest_v = dest.get(key, {})
            _deep_update(src=src_v, dest=dest_v)
            dest[key] = dest_v
        else:
            dest[key] = src_v


def _init_meta(file_metas: typ.List[Metadata]) -> Metadata:
    build_tt = time.localtime()
    meta: Metadata = {
        'litprog_version'   : __version__,
        'build_timestamp'   : time.strftime("%a %Y-%m-%d %H:%M:%S %Z", build_tt),
        'vcs_name'          : "",
        'vcs_revision'      : "",
        'vcs_revision_short': "",
        'vcs_dirty_files'   : set(),
        'configuration'     : {
            'concurrent': True,
        },
    }

    for file_meta in file_metas:
        if file_meta:
            _deep_update(src=file_meta, dest=meta)

    author_aliases = {}
    for raw_alias in meta.get('vcs_author_alias', []):
        alias_re = re.compile(raw_alias['alias_regex'])
        author_aliases[alias_re] = (raw_alias['real_name'], raw_alias['real_email'])

    vcs_api = vcs.get_vcs_api()
    if vcs_api:
        head_rev      = vcs_api.head_rev()
        commits       = vcs_api.commits()
        stats_by_file = vcs_timeline.stats_by_file(commits, author_aliases)
        overall_stats = vcs_timeline.calc_overall_stats(commits, author_aliases)
        meta.update(
            {
                'overall_stats'     : overall_stats,
                'stats_by_file'     : stats_by_file,
                'vcs_name'          : vcs_api.name,
                'vcs_revision'      : head_rev,
                'vcs_revision_short': head_rev[:10],
                'vcs_dirty_files'   : set(vcs_api.status()),
            }
        )

    return meta


StaticPaths = typ.Set[typ.Tuple[str, str]]


HEADLINE_RE = re.compile('<h1 id="([^"]+)">')


def _iter_file_nav_htmls(file_items: typ.List[FileItem]) -> typ.Iterable[str]:
    top_level_toc = []
    for file_item in file_items:
        for toc in file_item.html_res.toc_tokens:
            if toc['level'] == 1:
                filename_md   = str(file_item.md_path.name)
                filename_html = filename_md[: -len(file_item.md_path.suffix)] + ".html"
                file_nav_id   = filename_html + "#" + toc['id']
                top_level_toc.append((file_nav_id, toc['id'], toc['name']))

    for file_item in file_items:
        content_html = file_item.html_res.raw_html
        headline     = HEADLINE_RE.search(content_html)
        if headline is None:
            continue

        cur_nav_id = headline.group(1)

        nav_html_parts = ['<div class="toc"><ul>']
        for file_nav_id, nav_id, nav_name in top_level_toc:
            if nav_id == cur_nav_id:
                subnav_html = file_item.html_res.toc_html
                l_index     = subnav_html.find("<ul>")
                r_index     = subnav_html.rfind("</ul>")
                subnav_html = subnav_html[l_index + 4 : r_index]
                html_part   = f"<li><ul>{subnav_html}</ul></li>"
                html_part   = html_part.replace("#" + nav_id, file_nav_id, 1)
                nav_html_parts.append(html_part)
            else:
                nav_html_parts.append(f'<li><a href="{file_nav_id}">{nav_name}</a></li>')
        nav_html_parts.append("</ul></div>")

        nav_html = "\n".join(nav_html_parts)

        has_footnotes = bool(FOOTNOTES_RE.search(content_html))
        nav_html      = html_postproc.postproc_nav_html(nav_html, has_footnotes)

        yield nav_html


def _write_screen_html(file_items: typ.List[FileItem], html_dir: pl.Path) -> None:
    inital_url = ""

    nav_htmls = list(_iter_file_nav_htmls(file_items))
    for nav_html, (md_path, meta, html_res, block_linenos) in zip(nav_htmls, file_items):
        html_fname = md_path.stem + ".html"
        html_fpath = html_dir / html_fname
        logger.info(f"writing '{md_path}' -> '{html_fpath}'")
        htmls        = html_postproc.postproc4screen(html_res, block_linenos, nav_html)
        wrapped_html = wrap_content_html(htmls.content, 'screen', meta, htmls)
        with html_fpath.open(mode="w") as fobj:
            fobj.write(wrapped_html)

        if inital_url:
            inital_url = min(inital_url, html_fname)
        else:
            inital_url = html_fname

    if inital_url:
        with (html_dir / "index.html").open(mode="w") as fobj:
            fobj.write(INDEX_HTML.format(inital_url))


Package = typ.NewType('Package', str)


def _iter_package_paths() -> typ.Iterator[tuple[Package, str]]:
    available_filepaths = {
        package: list(importlib_resources.contents(package))
        for package in ["litprog.static", "litprog.static.fonts"]
    }

    for static_fpath in SELECTED_STATIC_DEPS:
        dirpath, fname = static_fpath.rsplit("/", 1)
        package = Package("litprog." + dirpath.replace("/", "."))

        pkg_fname_re = re.compile(fname)
        for pkg_fname in available_filepaths[package]:
            if pkg_fname_re.match(pkg_fname):
                yield package, pkg_fname


def _write_static_files(captured_static_paths: StaticPaths, html_dir: pl.Path) -> None:
    # copy/update static dependencies references in markdown files
    for src_path_str, tgt_path_str in sorted(captured_static_paths):
        # logger.info(f"copy {src_path_str} -> {tgt_path_str}")
        shutil.copy(src_path_str, tgt_path_str)

    out_static_dir = html_dir / "static"
    out_static_dir.mkdir(parents=True, exist_ok=True)

    for package, pkg_fname in _iter_package_paths():
        out_fpath = out_static_dir / pkg_fname
        with importlib_resources.path(package, pkg_fname) as in_path:
            if out_fpath.exists() and in_path.stat().st_mtime < out_fpath.stat().st_mtime:
                continue
            logger.debug(f"copy {in_path} -> {out_fpath}")
            with in_path.open(mode="rb") as in_fobj:
                with out_fpath.open(mode="wb") as out_fobj:
                    shutil.copyfileobj(in_fobj, out_fobj)


def gen_html(ctx: parse.Context, html_dir: pl.Path) -> None:
    logger.info(f"Writing html to '{html_dir}'")
    if not html_dir.exists():
        html_dir.mkdir(parents=True)

    md_file: parse.MarkdownFile

    file_metas = [md_file.parse_front_matter_meta() for md_file in ctx.files]
    cur_meta   = _init_meta(file_metas)

    captured_static_paths: StaticPaths = set()

    file_items: typ.List[FileItem] = []
    for file_meta, md_file in zip(file_metas, ctx.files):
        logger.info(f"processing '{md_file.md_path}'")
        md_text: MarkdownText = str(md_file)
        cur_meta.update(file_meta)
        md_filepath = str(md_file.md_path)
        cur_meta['md_filepath'  ] = md_filepath
        cur_meta['is_file_dirty'] = md_filepath in cur_meta['vcs_dirty_files']

        if cur_meta.get('stats_by_file') and md_filepath in cur_meta['stats_by_file']:
            base_name, _ = re.subn(r"[^\w]", "_", str(md_file.md_path.name))
            svg_name = base_name + "_authors_timeline.svg"
            svg_path = html_dir / svg_name
            stats    = cur_meta['stats_by_file'][md_filepath]
            vcs_timeline.write_stats_svg(stats, svg_path)
            cur_meta['authors_timeline_svg'] = svg_name

        for img_tag in md_file.image_tags():
            src_path = md_file.md_path.parent / img_tag.url
            if src_path.exists():
                tgt_path = html_dir / img_tag.url
                captured_static_paths.add((str(src_path), str(tgt_path)))

        html_res: md2html.HTMLResult = md2html.md2html(md_text, filename=str(md_file.md_path))

        if html_res.raw_html:
            block_linenos = list(md_file.iter_block_linenos())
            file_item     = FileItem(
                md_file.md_path,
                cur_meta.copy(),
                html_res,
                block_linenos,
            )
            file_items.append(file_item)

    _write_screen_html(file_items, html_dir)
    _write_static_files(captured_static_paths, html_dir)


def gen_pdf(
    ctx     : parse.Context,
    html_dir: pl.Path,
    pdf_dir : pl.Path,
    formats : typ.Sequence[str] = PRINT_FORMATS,
) -> None:
    if not pdf_dir.exists():
        pdf_dir.mkdir(parents=True)

    # TODO (mb 2021-01-21): cover page with title and author(s)
    #   timeline for whole project

    file_metas = [md_file.parse_front_matter_meta() for md_file in ctx.files]
    cur_meta   = _init_meta(file_metas)

    all_md_texts = []
    block_linenos: typ.List[typ.Tuple[int, int]] = []
    for file_meta, md_file in zip(file_metas, ctx.files):
        md_text: MarkdownText = str(md_file)
        cur_meta.update(file_meta)
        all_md_texts.append(md_text)
        block_linenos.extend(md_file.iter_block_linenos())

    full_md_text = "\n\n".join(all_md_texts)

    html_res: md2html.HTMLResult = md2html.md2html(full_md_text)
    multipage_formats = {fmt for fmt in formats if fmt in MULTIPAGE_FORMATS}
    onepage_formats   = set(formats) - set(multipage_formats)
    for fmt in multipage_formats:
        part_page_fmt = MULTIPAGE_FORMATS[fmt]
        onepage_formats.add(part_page_fmt)

    for fmt in onepage_formats:
        print_html   = html_postproc.postproc4print(html_res, fmt, block_linenos)
        wrapped_html = wrap_content_html(print_html, fmt, cur_meta)
        html_fpath   = pdf_dir / (fmt + ".html")
        pdf_fpath    = pdf_dir / (fmt + ".pdf")
        with html_fpath.open(mode="w") as fobj:
            fobj.write(wrapped_html)

        logger.info(f"converting '{html_fpath}' -> '{pdf_fpath}'")
        html2pdf.html2pdf(wrapped_html, pdf_fpath, html_dir)

    for fmt in multipage_formats:
        part_page_fmt       = MULTIPAGE_FORMATS[fmt]
        part_page_pdf_fpath = pdf_dir / (part_page_fmt + ".pdf")
        booklet_pdf_fpath   = pdf_dir / (fmt           + ".pdf")
        logger.info(f"creating booklet '{part_page_pdf_fpath}' -> '{booklet_pdf_fpath}'")
        pdf_booklet.create(in_path=part_page_pdf_fpath, out_path=booklet_pdf_fpath)



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import os
import re
import sys
import math
import typing as typ
import pathlib as pl
import datetime as dt
import textwrap
from xml.sax.saxutils import escape as xml_escape

from . import vcs

Filename      = str
AuthorName    = str
Timeline      = typ.List[int]
Contributions = typ.Dict[AuthorName, Timeline]

FileStats = typ.Dict[Filename, Contributions]

AuthorFull    = typ.Tuple[str, str]
AuthorAliases = typ.Dict[typ.Pattern, AuthorFull]


class Stats(typ.NamedTuple):
    timeline_start: str
    timeline_end  : str
    filestats     : FileStats


StatsByAuthor = typ.Dict[AuthorFull, typ.List[typ.Tuple[dt.date, int, int]]]
StatsByFile   = typ.Dict[str       , StatsByAuthor]


def stats_by_file(commits: typ.List[vcs.Commit], author_aliases: AuthorAliases) -> StatsByFile:
    stats: StatsByFile = {}
    for commit in commits:
        name                = commit.author.strip()
        email               = commit.email.strip()
        author              = (name, email)
        _maybe_author_alias = f"{name} <{email}>"
        for alias_re, author_normalized in author_aliases.items():
            if alias_re.match(_maybe_author_alias):
                author = author_normalized

        for change in commit.changes:
            total_rel = change.added_rel + change.deleted_rel
            if total_rel > 0:
                added_ratio   = change.added_rel   / total_rel
                deleted_ratio = change.deleted_rel / total_rel
                added_lines   = round(change.changed_lines * added_ratio)
                deleted_lines = round(change.changed_lines * deleted_ratio)
            else:
                added_lines   = 0
                deleted_lines = 0

            if change.filename in stats:
                file_stats = stats[change.filename]
            else:
                file_stats = stats[change.filename] = {}

            date      = dt.date(*map(int, commit.date.split("-")))
            file_stat = (date, added_lines, deleted_lines)
            if author in file_stats:
                file_stats[author].append(file_stat)
            else:
                file_stats[author] = [file_stat]

    return stats


def calc_overall_stats(commits: typ.List[vcs.Commit], author_aliases: AuthorAliases) -> StatsByAuthor:
    stats        : StatsByFile   = stats_by_file(commits, author_aliases)
    overall_stats: StatsByAuthor = {}
    for file_stats_by_author in stats.values():
        for author, file_stats in file_stats_by_author.items():
            if author in overall_stats:
                overall_stats[author].extend(file_stats)
            else:
                overall_stats[author] = list(file_stats)
    return overall_stats


class Row(typ.NamedTuple):
    total : float
    author: str
    email : str
    scores: typ.List[float]


class Grid(typ.NamedTuple):
    min_date : dt.date
    max_date : dt.date
    day_width: float  # number of cells
    row_width: int
    rows     : typ.List[Row]


def calc_stats_grid(stats: StatsByAuthor, max_width: int) -> Grid:
    dates = {
        date for file_stats in stats.values() for date, added, removed in file_stats if added + removed > 0
    }

    min_date   = min(dates)
    max_date   = max(dates)
    total_days = (max_date - min_date).days + 1
    row_width  = min(max_width, total_days)
    day_width  = row_width / total_days

    rows = []
    for (author, email), file_stats in stats.items():
        row_scores = [0.0] * row_width

        total_score = 0.0
        for date, added, removed in file_stats:
            days    = (date - min_date).days
            changed = added + removed
            i       = int(row_width * days / total_days)
            score   = math.log2(changed + 1)
            row_scores[i] += score
            total_score += score

        row = Row(total_score, author, email, row_scores)
        rows.append(row)

    # sort desc by "most" contributions
    rows.sort(reverse=True)
    return Grid(min_date, max_date, day_width, row_width, rows)


Legend = typ.List[typ.Tuple[int, str]]


def quarter(month: int) -> int:
    return math.ceil(month / 3)


def calc_legend(grid: Grid) -> Legend:
    num_days  = (grid.max_date - grid.min_date).days
    num_years = num_days / 365.25

    if num_years <= 4:
        year_step  = 0
        month_step = 3
    elif num_years <= 8:
        year_step  = 1
        month_step = 0
    else:
        year_step  = int(num_years / 5)
        month_step = 0

    cur_year  = grid.min_date.year
    cur_month = quarter(grid.min_date.month) * 3 - 2

    while dt.date(cur_year, cur_month, 1) <= grid.min_date:
        cur_year  += year_step
        cur_month += month_step
        if cur_month > 12:
            cur_year += 1
            cur_month = 1

    prev_date = grid.min_date

    max_date = (grid.max_date.year + 1, grid.max_date.month + 3)
    widths   = []
    yyyy_qtr = []
    while (cur_year, cur_month) < max_date:
        cur_date = dt.date(cur_year, cur_month, 1)
        width    = (cur_date - prev_date).days
        widths.append(width)
        qtr = quarter(prev_date.month)
        yyyy_qtr.append((prev_date.year, qtr))

        prev_date = cur_date
        cur_year  += year_step
        cur_month += month_step
        if cur_month > 12:
            cur_year += 1
            cur_month = 1

    legend = []
    for i, ((year, qtr), width) in enumerate(zip(yyyy_qtr, widths)):
        cell_width = round(width * grid.day_width)
        if i == 0:
            legend.append((cell_width, ""))
        elif qtr == 1:
            legend.append((cell_width, "|" + str(year)))
        else:
            legend.append((cell_width, "╷Q" + str(qtr)))

    return legend


def max_score(rows: typ.List[Row]) -> float:
    return max(sum((row.scores for row in rows), []))


def grid_to_text(grid: Grid):
    # legend
    legend = calc_legend(grid)

    timescale_parts = [part_str.ljust(part_width) for part_width, part_str in legend]

    legend_str = "".join(timescale_parts)
    legend_str = legend_str[: grid.row_width]
    if len(legend_str) > 6:
        legend_str = legend_str[:-6] + (legend_str[-6:].rstrip("Q0123456789 "))
    legend_str = legend_str.ljust(grid.row_width)

    min_date_str = grid.min_date.isoformat()
    max_date_str = grid.max_date.isoformat()

    header_line = f" {min_date_str:>55} |" + legend_str + "| " + max_date_str

    # timelines
    chars = " .:!#"
    chars = " ▁▃▄▅▆▇█"
    chars = " ⡀⣀⣄⣤⣦⣶⣷⣿"

    scale = (len(chars) - 1) / max_score(grid.rows)

    lines = [header_line]
    for total_score, author, email, row_scores in grid.rows:
        timeline = "".join(chars[round(s * scale)] for s in row_scores)
        lines.append(f"{author:<25} {email:<30} |{timeline}| {total_score:7.2f}")

    return "\n".join(lines)


def _iter_marker_elems(grid: Grid, text_width: int) -> typ.Iterable[str]:
    legend = calc_legend(grid)

    max_date_x_offset = text_width + grid.row_width
    x_offset          = text_width
    for part_width, part_str in legend:
        legend_elem_str = f"""
        <text x="0" y="0" width="{part_width}">
            <tspan x="{x_offset}" dy="16">{xml_escape(part_str)}</tspan>
        </text>
        """
        x_offset += part_width
        yield legend_elem_str
        if x_offset + 20 > max_date_x_offset:
            return


def _iter_legend_elems(grid: Grid, text_width: int) -> typ.Iterable[str]:
    max_date_x_offset = text_width + grid.row_width
    min_date_x_offset = text_width - 80

    yield f"""
    <text x="0" y="0" width="80">
        <tspan x="{min_date_x_offset}" dy="16">{grid.min_date.isoformat()}|</tspan>
    </text>
    """

    yield from _iter_marker_elems(grid, text_width)

    yield f"""
    <text x="0" y="0" width="80">
        <tspan x="{max_date_x_offset}" dy="16">|{grid.max_date.isoformat()}</tspan>
    </text>
    """


class TextRow(typ.NamedTuple):
    author  : str
    email   : str
    y_offset: int
    scores  : typ.List[float]


def _iter_image_elems(
    grid      : Grid,
    rows      : typ.List[TextRow],
    text_width: int,
    img_width : int,
    max_height: int,
    padding   : int,
) -> typ.Iterable[str]:
    scale = max_height / max_score(grid.rows)

    for i, row in enumerate(rows):
        text_elem_str = f"""
        <text x="0" y="{row.y_offset}" width="{text_width}" height="{max_height}">
            <tspan x="4" dy="14">{xml_escape(row.author)}</tspan>
            <tspan x="4" dy="16">{xml_escape(row.email)}</tspan>
        </text>
        <rect class="bg" x="{text_width}" y="{row.y_offset}" width="{grid.row_width}" height="{max_height}"/>
        """

        x_offset   = text_width
        base_width = max(1, int((grid.row_width - 5) / len(row.scores)))

        timeline_elems = []
        for score in row.scores:
            width = base_width * 5
            if score:
                height = max(3, round(scale * score))
                elem_y = row.y_offset + max_height - height
                elem_x = x_offset - width
                elem   = f'<rect x="{elem_x}" y="{elem_y}" width="{width}" height="{height}" fill="#0009"/>'
                timeline_elems.append(elem.strip())
            x_offset += base_width

        timeline_elems_str = "\n".join(timeline_elems)

        row_str = textwrap.dedent(text_elem_str) + textwrap.dedent(timeline_elems_str)
        yield row_str.strip()

        if i + 1 < len(rows):
            line_y = round(row.y_offset + max_height + padding / 2)
            line   = f'<line x1="4" x2="{img_width - 8}" y1="{line_y}" y2="{line_y}" stroke="#888"></line>'
            yield line


SVG_IMG_TMPL = """
<svg xmlns="http://www.w3.org/2000/svg" width="{img_width}" height="{img_height}" version="1.1">
  <defs>
    <style type='text/css'><![CDATA[
      text, tspan {{
        font-size: 14px;
        font-family: lp-iosevka, 'Iosevka Term SS05', monospace;
        font-weight: normal;
      }}
      rect.bg {{
        fill: #F00B;
        display: none;
      }}
    ]]></style>
  </defs>
  {img_content}
</svg>
"""


def grid_to_svg(grid: Grid):
    padding       = 5
    max_height    = 35
    row_height    = max_height + padding
    legend_height = row_height // 2
    img_height    = legend_height + row_height * len(grid.rows)

    textwidth_chars = 19
    text_rows: typ.List[TextRow] = []
    for i, (_, author, email, row_scores) in enumerate(grid.rows):
        y_offset = i * row_height + 4 + legend_height
        email    = ("<" + email + ">").lower()

        textwidth_chars = max(textwidth_chars, len(author), len(email))
        text_rows.append(TextRow(author, email, y_offset, row_scores))

    text_width = int(textwidth_chars * 8 + padding)
    img_width  = text_width + grid.row_width + 80

    legend_elems = list(_iter_legend_elems(grid, text_width))
    image_elems  = list(_iter_image_elems(grid, text_rows, text_width, img_width, max_height, padding))

    svg_img = SVG_IMG_TMPL.format(
        img_width=img_width,
        img_height=img_height,
        img_content="\n".join(legend_elems + image_elems),
    )
    return textwrap.dedent(svg_img).strip()


def write_stats_svg(stats: StatsByAuthor, svg_path: pl.Path) -> None:
    grid = calc_stats_grid(stats, max_width=120)
    # print()
    # print(grid_to_text(grid))

    grid    = calc_stats_grid(stats, max_width=400)
    svg_img = grid_to_svg(grid)
    with svg_path.open(mode="w", encoding="utf-8") as fobj:
        fobj.write(svg_img)


PROJECT_NAMES = [
    # "pycalver",
    # "bootstrapit",
    "markdown-katex",
    # "markdown-svgbob",
    "markdown-aafigure",
    "lib3to6",
    "litprog",
    # "lexid",
    "imeig",
    "sbk",
]


def main() -> int:
    cwd            = os.getcwd()
    mb_author      = ("Manuel Barkhau", "mbarkhau@gmail.com")
    author_aliases = {
        re.compile(r".*barkhau.*"     , re.IGNORECASE): mb_author,
        re.compile(r".*root@vserver.*", re.IGNORECASE): mb_author,
    }
    for project_name in PROJECT_NAMES:
        project_path = f"../{project_name}/"
        os.chdir(project_path)

        vcs_api = vcs.get_vcs_api()
        assert vcs_api is not None
        commits       = vcs_api.commits()
        overall_stats = calc_overall_stats(commits, author_aliases)

        commits       = vcs_api.commits()
        overall_stats = calc_overall_stats(commits, author_aliases)
        os.chdir(cwd)

        stats_img_path = pl.Path("scratch", project_name + ".svg")
        write_stats_svg(overall_stats, stats_img_path)
    return 0


if __name__ == '__main__':
    import pretty_traceback

    pretty_traceback.install(envvar='ENABLE_PRETTY_TRACEBACK')
    sys.exit(main())



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import os
import json
import time
import shlex
import typing as typ
import logging
import os.path
import pathlib as pl
import threading
import subprocess as sp

logger = logging.getLogger(__name__)

InputPaths = typ.Sequence[str]
FilePaths  = typ.Iterable[pl.Path]

ExitCode = int


class RawCapturedLine(typ.NamedTuple):
    ts  : float
    line: str


class CapturedLine(typ.NamedTuple):
    ts    : float
    line  : str
    is_err: bool


class Capture(typ.NamedTuple):
    command    : str
    exit_status: int
    runtime    : float
    lines      : typ.List[CapturedLine]


CaptureData = bytes


def loads_capture(capture_bytes: CaptureData) -> Capture:
    capture_json = capture_bytes.decode("utf-8")
    capture_obj  = json.loads(capture_json)
    command, exit_status, runtime, lines_args = capture_obj

    lines = [CapturedLine(*line_args) for line_args in lines_args]
    return Capture(
        command,
        exit_status,
        runtime,
        lines,
    )


def dumps_capture(capture: Capture) -> CaptureData:
    line_args   = [[line.ts, line.line, line.is_err] for line in capture.lines]
    capture_obj = [
        capture.command,
        capture.exit_status,
        capture.runtime,
        line_args,
    ]
    capture_json  = json.dumps(capture_obj)
    capture_bytes = capture_json.encode("utf-8")
    return capture_bytes


class SessionException(Exception):
    pass


Environ = typ.Mapping[str, str]
# make mypy fail if this is the wrong type
_: Environ = os.environ


def _gen_captured_lines(
    raw_lines: typ.Iterable[bytes], encoding: str = "utf-8", debug_log: bool = False
) -> typ.Iterable[RawCapturedLine]:
    for raw_line in raw_lines:
        # get timestamp as fast as possible after
        #   output was read
        ts = time.time()

        line_value = raw_line.decode(encoding)
        if debug_log:
            logger.debug(f"read {len(raw_line)} bytes")
        yield RawCapturedLine(ts, line_value)


def _read_loop(
    sp_output_pipe: typ.IO[bytes],
    captured_lines: typ.List[RawCapturedLine],
    encoding      : str = "utf-8",
) -> None:
    raw_lines = iter(sp_output_pipe.readline, b'')
    for captured_line in _gen_captured_lines(raw_lines, encoding=encoding):
        captured_lines.append(captured_line)


class CapturingThread(typ.NamedTuple):
    thread: threading.Thread
    lines : typ.List[RawCapturedLine]


def _start_reader(
    sp_output_pipe: typ.IO[bytes],
    encoding      : str = "utf-8",
) -> CapturingThread:
    captured_lines: typ.List[RawCapturedLine] = []
    read_loop_thread = threading.Thread(target=_read_loop, args=(sp_output_pipe, captured_lines, encoding))
    read_loop_thread.start()
    return CapturingThread(read_loop_thread, captured_lines)


AnyCommand = typ.Union[str, typ.List[str]]


def _normalize_command(command: AnyCommand) -> typ.List[str]:
    if isinstance(command, str):
        return shlex.split(command)
    elif isinstance(command, list):
        return command
    else:
        err_msg = f"Invalid command: {command}"
        raise Exception(err_msg)


class InteractiveSession:

    encoding: str
    start   : float
    end     : float

    _retcode: typ.Optional[int]
    _proc   : sp.Popen
    _stdin  : typ.Optional[typ.IO[bytes]]
    _stdout : typ.IO[bytes]
    _stderr : typ.IO[bytes]

    _in_cl : typ.List[RawCapturedLine]
    _out_ct: CapturingThread
    _err_ct: CapturingThread

    def __init__(
        self,
        cmd: AnyCommand,
        *,
        env      : typ.Optional[Environ] = None,
        encoding : str  = "utf-8",
        debug_log: bool = False,
    ) -> None:
        _env: Environ
        if env is None:
            _env = os.environ.copy()
        else:
            _env = env

        self.encoding  = encoding
        self.debug_log = debug_log
        self.start     = time.time()
        self.end       = -1.0
        self._retcode  = None

        cmd_parts = _normalize_command(cmd)
        if self.debug_log:
            logger.debug(f"popen {cmd_parts}")
        self._proc = sp.Popen(cmd_parts, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE, env=_env)

        assert self._proc.stdin  is not None
        assert self._proc.stdout is not None
        assert self._proc.stderr is not None
        self._stdin  = self._proc.stdin
        self._stdout = self._proc.stdout
        self._stderr = self._proc.stderr

        _enc = encoding

        self._in_cl  = []
        self._out_ct = _start_reader(self._stdout, _enc)
        self._err_ct = _start_reader(self._stderr, _enc)

    def send(self, input_str: str, delay: float = 0) -> None:
        self._in_cl.append(RawCapturedLine(time.time(), input_str))
        input_data = input_str.encode(self.encoding)

        _stdin = self._stdin
        if _stdin:
            _stdin.write(input_data)

            if delay:
                _stdin.flush()
        if delay:
            time.sleep(delay)

    @property
    def retcode(self) -> int:
        _stdin = self._stdin
        if _stdin:
            _stdin.flush()
        return self.wait()

    def _assert_retcode(self) -> None:
        if self._retcode is None:
            cls = type(self)
            msg = f"'{cls}.wait()' must be called before accessing captured output."
            raise RuntimeError(msg)

    def _wait(self, timeout) -> int:
        try:
            max_time = self.start + timeout
            while True:
                time_left = max_time - time.time()
                if self.debug_log:
                    logger.debug(f"poll {time_left}")
                time.sleep(min(0.01, max(0, time_left)))
                returncode = self._proc.poll()
                if returncode is not None:
                    if self.debug_log:
                        logger.debug(f"poll() returned {returncode}")
                    break
                if time.time() > max_time:
                    if self.debug_log:
                        logger.debug("timeout")
                    break
        finally:
            if returncode is None:
                if self.debug_log:
                    logger.debug("sending SIGTERM")
                self._proc.terminate()
                returncode = self._proc.wait()

        assert returncode is not None
        self._retcode = returncode
        self.end      = time.time()
        return returncode

    def wait(self, timeout=1) -> int:
        if self._retcode is not None:
            return self._retcode

        if self.debug_log:
            logger.debug(f"wait with timeout={timeout}")
        returncode: typ.Optional[int] = None
        _stdin = self._stdin
        if _stdin:
            try:
                _stdin.close()
            except BrokenPipeError:
                # NOTE (mb 2020-06-05): Some subprocesses exit so fast that the pipe
                #   may already be closed.
                logger.debug("stdin already closed")

        returncode = self._wait(timeout)
        self._out_ct.thread.join()
        self._err_ct.thread.join()
        return returncode

    @property
    def out_lines(self) -> typ.List[RawCapturedLine]:
        return self._out_ct.lines

    @property
    def err_lines(self) -> typ.List[RawCapturedLine]:
        return self._err_ct.lines

    def iter_lines(self) -> typ.Iterable[CapturedLine]:
        out_lines = iter(self.out_lines)
        err_lines = iter(self.err_lines)
        out_line  = next(out_lines, None)
        err_line  = next(err_lines, None)

        while True:
            if out_line and err_line:
                if out_line.ts <= err_line.ts:
                    yield CapturedLine(out_line.ts, out_line.line, False)
                    out_line = next(out_lines, None)
                else:
                    yield CapturedLine(err_line.ts, err_line.line, True)
                    err_line = next(err_lines, None)
            elif err_line:
                yield CapturedLine(err_line.ts, err_line.line, True)
                err_line = next(err_lines, None)
            elif out_line:
                yield CapturedLine(out_line.ts, out_line.line, False)
                out_line = next(out_lines, None)
            else:
                break

    def output_lines(self) -> typ.List[CapturedLine]:
        return list(self.iter_lines())

    def iter_stdout(self) -> typ.Iterable[str]:
        for _ts, line in self._out_ct.lines:
            yield line

    def iter_stderr(self) -> typ.Iterable[str]:
        for _ts, line in self._err_ct.lines:
            yield line

    def __iter__(self) -> typ.Iterable[str]:
        all_lines = self._in_cl + self._out_ct.lines + self._err_ct.lines
        for captured_line in sorted(all_lines):
            yield captured_line.line

    @property
    def runtime(self) -> float:
        self._assert_retcode()
        return self.end - self.start

    @property
    def stdout(self) -> str:
        return "".join(self.iter_stdout())

    @property
    def stderr(self) -> str:
        return "".join(self.iter_stderr())


class DebugInteractiveSession(InteractiveSession):

    start: float
    end  : float

    _retcode: typ.Optional[int]
    _proc   : sp.Popen
    _stdin  : typ.Optional[typ.IO[bytes]]

    _in_cl: typ.List[RawCapturedLine]

    def __init__(
        self,
        cmd: AnyCommand,
        *,
        env      : typ.Optional[Environ] = None,
        encoding : str  = "utf-8",
        debug_log: bool = False,
    ) -> None:
        # pylint: disable=super-init-not-called; initialization is substantially different
        _env: Environ
        if env is None:
            _env = os.environ.copy()
        else:
            _env = env

        self.encoding  = encoding
        self.debug_log = debug_log
        self.start     = time.time()
        self.end       = -1.0
        self._retcode  = None

        cmd_parts = _normalize_command(cmd)
        if self.debug_log:
            logger.debug(f"popen {cmd_parts}")
        self._proc = sp.Popen(cmd_parts, env=_env)

        self._stdin = self._proc.stdin
        self._in_cl = []

    def wait(self, timeout=1) -> int:
        if self._retcode is not None:
            return self._retcode

        return self._wait(timeout)

    def output_lines(self) -> typ.List[CapturedLine]:
        return []



# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

# After reading "https://apenwarr.ca/log/20181113" I've
# concluded, that we need to
#  1. create an index of files we care about, with
#   the os.stat result and content hash for every path
#  2. recalculate the content hash if
#   - anything from os.stat changes
#   - mtime is zero
#   - mtime is between now - estimated precision
# The estimated precision is based on the number of
# trailing zeros on the mtime.

# files that are dependencies must have an mtime
#   that is newer than the start time of the build.
#   In other words, dependencies are only considered
#   to be satisfied, if they were created/updated
#   after the build started.
#
#   https://docs.microsoft.com/en-us/windows/desktop/sysinfo/file-times
#   > create time on FAT is 10 milliseconds,
#   > while write time has a resolution of 2 seconds
#
# Better approach:
# - https://github.com/ziglang/zig/issues/2045

import os
import sys
import json
import time
import uuid
import socket
import typing as typ
import hashlib
import logging
import pathlib as pl

logger = logging.getLogger(__name__)


def _machine_id() -> str:
    """Generate machine specific id.

    The index is only valid if it was written by the same machine
    that it was generated on.

    An index might be shared if multiple builds run over an
    NFS file system. Raise an error if this is detected and
    include the path to the index file in the error message,
    so the user can decide if they want to delete it and
    rebuild from scratch.
    """
    node = uuid.getnode()
    if node & 0b10000000 > 0:
        # If all attempts to obtain the hardware address fail, we
        # choose a random 48-bit number with its eighth bit set to 1
        # as recommended in RFC 4122.
        node = 0

    host = socket.gethostname()
    return f"{node}_{host}"


IS_BLAKE2_AVAILABLE = False

try:
    hashlib.new('blake2b')
except ValueError as ex:
    assert "unsupported hash type blake2b" in str(ex)
    IS_BLAKE2_AVAILABLE = False


HexDigest = str


def _file_digest(path: pl.Path) -> HexDigest:
    # https://blake2.net/
    if IS_BLAKE2_AVAILABLE:
        id_sum = hashlib.new('blake2b')
    else:
        id_sum = hashlib.new('sha1')

    with path.open(mode="rb") as fobj:
        while True:
            chunk = fobj.read(65536)
            if chunk:
                id_sum.update(chunk)
            else:
                break
    return id_sum.hexdigest()


class CheckResult(typ.NamedTuple):
    # NOTE: If the digest is set, it is
    #   the fresh from disk, not from the cache.
    #   This is done so we don't have to
    #   calculate the digest twice; once when
    #   checking if an entry is valid, and once
    #   again when updating the entry.
    ok    : bool
    digest: HexDigest


EntryKey = str
Target   = str


def entry_key(path: pl.Path) -> EntryKey:
    return str(path.absolute())


# NOTE: mypy doesn't like the stat_result type :-/, so we
#   convert it to something it knows how to deal with.


class Stat(typ.NamedTuple):

    mode : int
    ino  : int
    dev  : int
    nlink: int
    uid  : int
    gid  : int
    size : int
    atime: float
    mtime: float
    ctime: float


def mk_stat(path: pl.Path) -> Stat:
    stat_result = path.stat()

    # NOTE: getmtime is more precise than stat_result.st_mtime
    mtime: float = os.path.getmtime(str(path))

    return Stat(
        stat_result.st_mode,
        stat_result.st_ino,
        stat_result.st_dev,
        stat_result.st_nlink,
        stat_result.st_uid,
        stat_result.st_gid,
        stat_result.st_size,
        stat_result.st_atime,
        mtime,
        stat_result.st_ctime,
    )


class IndexEntry(typ.NamedTuple):

    path  : pl.Path
    stat  : Stat
    digest: HexDigest


MaybeEntry        = typ.Optional[IndexEntry]
EntryByKey        = typ.Dict[EntryKey, MaybeEntry]
EntryKeysByTarget = typ.Dict[Target  , typ.Set[EntryKey]]


def make_entry(path: pl.Path, check: CheckResult) -> MaybeEntry:
    if path.exists():
        digest_val = check.digest or _file_digest(path)
        return IndexEntry(path=path, stat=mk_stat(path), digest=digest_val)
    else:
        return None


class Index:

    machine_id: str
    index_file: pl.Path
    # The index_stat serves two purposes:
    #   1. To detect a concurrent build
    #   2. To invalidate cached digest of recently
    #       updated files.
    index_stat: Stat

    entries: EntryByKey
    targets: EntryKeysByTarget

    def __init__(self, index_file: pl.Path) -> None:
        self.index_file = index_file
        self.machine_id = _machine_id()
        self.entries    = {}
        self.targets    = {}

        try:
            if self.index_file.exists():
                self.load_index()
        except (IOError, ValueError, KeyError):
            warn_msg = f"Ignoring invalid/corrupted index file " f"'{self.index_file}'"
            logger.warning(warn_msg, exc_info=True)

        self.index_file.touch()
        self.index_stat = mk_stat(self.index_file)

    def load_index(self) -> None:
        with self.index_file.open(mode="rb") as fobj:
            data = json.loads(fobj.read().decode('utf-8'))

        if data['machine_id'] != self.machine_id:
            err_msg = (
                f"Index file '{self.index_file}' possibly "
                "created on different machine. Aborting to"
                "avoid concurrent build from different hosts."
                "If you know this is not an issue, delete "
                "the file and try again."
            )
            raise ValueError(err_msg)

        entries: EntryByKey = {}
        for key, entry in data['entries'].items():
            if entry is None:
                entries[key] = entry
            else:
                entries[key] = IndexEntry(pl.Path(entry['path']), Stat(*entry['stat']), entry['digest'])

        targets: EntryKeysByTarget = {}
        for target, deps in data['targets'].items():
            targets[target] = set(deps)

        self.entries = entries
        self.targets = targets

    def dump_index(self) -> None:
        EntryData = typ.Optional[typ.Dict[str, typ.Any]]
        entries: typ.Dict[str, EntryData] = {}

        for key, entry in self.entries.items():
            if entry:
                entries[key] = {
                    'path'  : str(entry.path),
                    'stat'  : list(entry.stat),
                    'digest': entry.digest,
                }
            else:
                entries[key] = None

        targets = {target: list(deps) for target, deps in self.targets.items()}
        data    = {'machine_id': self.machine_id, 'entries': entries, 'targets': targets}

        nonce    = time.time()
        tmp_name = self.index_file.name + f".{nonce}.tmp"
        tmp_file = self.index_file.parent / tmp_name
        with tmp_file.open(mode="wb") as fobj:
            text = json.dumps(data, indent=2)
            fobj.write(text.encode("utf-8"))

        if self.has_index_changed():
            tmp_file.unlink()
            err_msg = "WARNING: Concurrent update of index detected. This may result in a corrupted build."
            raise Exception(err_msg)
        else:
            tmp_file.rename(self.index_file)

    def has_index_changed(self) -> bool:
        if not self.index_file.exists():
            return True
        elif self.index_stat != mk_stat(self.index_file):
            return True
        else:
            return False

    def add_files(self, paths: typ.Iterable[pl.Path]) -> None:
        for path in paths:
            self.add_file(path)

    def add_file(self, path: pl.Path) -> None:
        check = self.check_path(path)
        if check.ok:
            return

        entry = make_entry(path, check)
        self.entries[entry_key(path)] = entry

    def check_path(self, path: pl.Path) -> CheckResult:
        # NOTE: There could be different levels of
        #   dirtyness.
        #   strict: hash and mtime must be unchanged
        #   hash: hash must be unchanged
        if entry_key(path) not in self.entries:
            return CheckResult(False, "")

        entry = self.entries[entry_key(path)]
        if entry is None:
            return CheckResult(False, "")
        elif not entry.path.exists():
            return CheckResult(False, "")
        elif entry.stat != mk_stat(path):
            return CheckResult(False, "")
        elif entry.stat.mtime <= self.index_stat.mtime - 2:
            # If mtime is very new, then it can't be trusted.
            return CheckResult(True, entry.digest)
        else:
            # fallback to digest check
            new_digest = _file_digest(path)
            check_ok   = new_digest == entry.digest
            return CheckResult(check_ok, new_digest)

    def is_target_done(self, target: Target, deps: typ.Set[pl.Path]) -> bool:
        for dep in deps:
            digest = self.check_path(dep)
            if not digest.ok:
                return False

        new_deps  = {entry_key(dep) for dep in deps}
        prev_deps = self.targets.get(target)
        return new_deps == prev_deps

    def mark_target_done(self, target: Target, deps: typ.Set[pl.Path]) -> None:
        for dep in deps:
            self.add_file(dep)

        self.targets[target] = {entry_key(dep) for dep in deps}


def _iter_paths(dirpath: pl.Path) -> typ.Iterable[pl.Path]:
    for root, _dirnames, filenames in os.walk(dirpath):
        for filename in filenames:
            yield pl.Path(root) / filename


ExitStatus = int


def selftest() -> ExitStatus:
    dirpaths = [
        pl.Path("src"),
        pl.Path("src_v2"),
        pl.Path("lit_v2"),
        pl.Path("lit_v3"),
        pl.Path("fonts"),
    ]
    filepaths: typ.List[pl.Path] = []
    for dirpath in dirpaths:
        filepaths.extend(_iter_paths(dirpath))

    idx  = Index(pl.Path(".litprog.index"))
    deps = set(_iter_paths(pl.Path("lit_v3")))
    print("done?", idx.is_target_done('moep', deps))

    idx.mark_target_done('moep', deps)
    for dep in deps:
        dep.touch()
    print("done?", idx.is_target_done('moep', deps))
    assert not idx.is_target_done('moep', deps)
    idx.mark_target_done('moep', deps)

    idx.add_files(filepaths)
    idx.dump_index()
    return 0


if __name__ == '__main__':
    sys.exit(selftest())



#!/home/mbarkhau/miniconda3/envs/litprog_py37/bin/python
# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT
import io
import sys
import typing as typ
import pathlib as pl

import PyPDF2 as pdf

PT_PER_INCH = 72
PT_PER_MM   = PT_PER_INCH / 25.4


PAPER_FORMATS_PT = {
    'A6'    : (105 * PT_PER_MM  , 148 * PT_PER_MM),
    'A5'    : (148 * PT_PER_MM  , 210 * PT_PER_MM),
    'A4'    : (210 * PT_PER_MM  , 297 * PT_PER_MM),
    'A3'    : (297 * PT_PER_MM  , 420 * PT_PER_MM),
    'Letter': (8.5 * PT_PER_INCH,  11 * PT_PER_INCH),
}

Points = float


def parse_to_pt(size: str, w_pt: Points, h_pt: Points) -> typ.Tuple[Points, Points]:
    if size.endswith("%"):
        x_pt = (float(size[:-1]) / 100) * w_pt
        y_pt = (float(size[:-1]) / 100) * h_pt
        return (x_pt, y_pt)
    elif size.endswith("pt"):
        size_pt = float(size[:-2])
        return (size_pt, size_pt)
    elif size.endswith("mm"):
        size_pt = float(size[:-2]) * PT_PER_MM
        return (size_pt, size_pt)
    elif size.endswith("inch"):
        size_pt = float(size[:-4]) * PT_PER_INCH
        return (size_pt, size_pt)
    else:
        raise Exception(f"Invalid unit {size}. Valid units are pt, mm, %")


class OpCoordinantes(typ.NamedTuple):
    next_w_pt   : float
    next_h_pt   : float
    scale_factor: float
    translate_x : float
    translate_y : float


PDFDocument = typ.Any


def _calc_op_coords(pdf_doc: PDFDocument, op: str, raw_ops: typ.Iterator[str]) -> OpCoordinantes:
    curr_media_box = pdf_doc.getPage(0).mediaBox

    cur_w_pt = float(curr_media_box.getWidth())
    cur_h_pt = float(curr_media_box.getHeight())

    # initial defaults are modified/updated by ops
    next_w_pt    = cur_w_pt
    next_h_pt    = cur_h_pt
    translate_x  = translate_y = 0.0
    scale_factor = 1.0

    if op == 'crop':
        crop_side = next(raw_ops).lower()
        crop_arg  = next(raw_ops)

        crop_l = crop_r = crop_t = crop_b = 0.0

        crop_x_pt, crop_y_pt = parse_to_pt(crop_arg, cur_w_pt, cur_h_pt)
        if crop_side in ('l', 'left'):
            crop_l = crop_x_pt
        elif crop_side in ('r', 'right'):
            crop_r = crop_x_pt
        elif crop_side in ('t', 'top'):
            crop_t = crop_y_pt
        elif crop_side in ('b', 'bottom'):
            crop_b = crop_y_pt
        else:
            errmsg = f"Invalid side {crop_side}. Valid sides are left, right, top, bottom"
            raise Exception(errmsg)

        translate_x -= crop_l
        translate_y -= crop_b
        next_w_pt = cur_w_pt - crop_l - crop_r
        next_h_pt = cur_h_pt - crop_t - crop_b
    elif op == 'scale':
        scale_type = next(raw_ops)
        scale_arg  = next(raw_ops)
        if scale_type == 'by':
            scale_factor = float(scale_arg)
        elif scale_type == 'to':
            next_w_pt, next_h_pt = PAPER_FORMATS_PT[scale_arg]
            scale_w      = next_w_pt / cur_w_pt
            scale_h      = next_h_pt / cur_h_pt
            scale_factor = min(scale_w, scale_h)
        else:
            raise Exception(f"Invalid scale arg: {scale_type}")
    else:
        # TODO: resize, translate
        raise Exception(f"Invalid operation: {op}")

    next_w_pt = round(next_w_pt)
    next_h_pt = round(next_h_pt)

    print(f"{op} w:{next_w_pt} h:{next_h_pt} s:{scale_factor} x:{translate_x} y:{translate_y}")
    return OpCoordinantes(
        next_w_pt,
        next_h_pt,
        scale_factor,
        translate_x,
        translate_y,
    )


def main(args: typ.Sequence[str] = sys.argv[1:]) -> int:
    # pylint: disable=dangerous-default-value

    in_path  = pl.Path(args[0])
    out_path = pl.Path(args[1])
    raw_ops  = iter(args[2:])

    curr_fobj = in_path.open(mode="rb")

    for op in raw_ops:
        curr_pdf  = pdf.PdfFileReader(curr_fobj)
        op_coords = _calc_op_coords(curr_pdf, op, raw_ops)

        next_pdf = pdf.PdfFileWriter()
        for cur_page in curr_pdf.pages:
            next_page = next_pdf.addBlankPage(width=op_coords.next_w_pt, height=op_coords.next_h_pt)
            next_page.mergeScaledTranslatedPage(
                cur_page,
                op_coords.scale_factor,
                op_coords.translate_x,
                op_coords.translate_y,
                expand=False,
            )
            # next_page.compressContentStreams()

        next_fobj = io.BytesIO()
        next_pdf.write(next_fobj)
        next_fobj.seek(0)

        curr_fobj.close()
        curr_fobj = next_fobj

    with out_path.open(mode="wb") as out_fobj:
        out_fobj.write(curr_fobj.read())

    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python

# This file is part of the litprog project
# https://github.com/litprog/litprog
#
# Copyright (c) 2018-2021 Manuel Barkhau (mbarkhau@gmail.com) - MIT License
# SPDX-License-Identifier: MIT

import litprog.cli

if __name__ == '__main__':
    litprog.cli.cli()
